#include <iostream>
#include <vcg/complex/complex.h>

#include <wrap/io_trimesh/import.h>
#include <wrap/io_trimesh/export.h>
#include <vcg/complex/algorithms/update/color.h>

#include <vcg/space/index/kdtree/kdtree.h>
#include <vcg/space/index/grid_static_ptr.h>
#include <vcg/space/index/spatial_hashing.h>
#include <vcg/space/index/octree.h>

int num_test = 1000;
int kNearest = 256;
float queryDist = 0.0037;
float bboxratio = 1000.0f;

using namespace vcg;


class CVertex;
class CFace;
class CEdge;

class CUsedTypes	: public vcg::UsedTypes < vcg::Use< CVertex >::AsVertexType, vcg::Use< CFace >::AsFaceType>{};
class CVertex		: public vcg::Vertex < CUsedTypes, vcg::vertex::Coord3f, vcg::vertex::Normal3f, vcg::vertex::Qualityf, vcg::vertex::BitFlags, vcg::vertex::Color4b>{};
class CFace			: public vcg::Face < CUsedTypes, vcg::face::VertexRef>{};

class CMesh			: public vcg::tri::TriMesh < std::vector< CVertex >, std::vector< CFace > > {};

int elapsed(int t)
{
    return ((clock()-t)*1000.0)/CLOCKS_PER_SEC;
}

void ComputeDistanceBetweenMeshes(CMesh &mesh1, CMesh &mesh2)
{
    typedef vcg::GridStaticPtr<CMesh::FaceType, CMesh::FaceType::ScalarType> MeshGrid;
    MeshGrid uniformGrid;
    uniformGrid.Set(mesh1.face.begin(), mesh1.face.end());
    
    // Compute for each vertex of mesh2 
    // the distance to the closest vertex of mesh1
    for(auto vi=mesh2.vert.begin();vi!=mesh2.vert.end();++vi)
    {
        CMesh::ScalarType minDist = std::numeric_limits<CMesh::ScalarType>::max();
        CMesh::FaceType *f;
        CMesh::ScalarType dist;
        Point3f p = vi->P();
        Point3f closestPt;
        typedef tri::FaceTmark<CMesh> MarkerFace;
        MarkerFace mf;
        mf.SetMesh(&mesh2);
        face::PointDistanceBaseFunctor<float> PDistFunct;
        f = uniformGrid.GetClosest(PDistFunct,mf,p,minDist,dist,closestPt);
   
    }


}

void ComputeAverageVertexDistance(CMesh& mesh)
{
    std::cout << "==================================================="<< std::endl;
    std::cout << "KDTree" << std::endl;
    int t0=clock();
    // Construction of the kdTree
    vcg::ConstDataWrapper<CMesh::VertexType::CoordType> wrapperVcg(&mesh.vert[0].P(), mesh.vert.size(), size_t(mesh.vert[1].P().V()) - size_t(mesh.vert[0].P().V()));
    vcg::KdTree<CMesh::ScalarType> kdTreeVcg(wrapperVcg);
    std::cout << "Build: " << elapsed(t0) <<  " ms" << std::endl;
    int nn=1;
    // Computation of the average distance between points the mesh
    float mAveragePointSpacing = 0;
    t0=clock();

    for (size_t i = 0; i < mesh.vert.size(); i++)
    {
        vcg::KdTree<CMesh::ScalarType>::PriorityQueue queue;
        // find the k-nearest points
        kdTreeVcg.doQueryK(mesh.vert[i].cP(), 16, queue);
        float mAveragePointSpacing=0;
        // compute the average distance
        for (int j = 0; j < queue.getNofElements(); j++)
        {
            mAveragePointSpacing += Distance( mesh.vert[i].cP(), mesh.vert[queue.getIndex(j)].cP() );
        }
        
        mesh.vert[i].Q() = mAveragePointSpacing / queue.getNofElements();    
    }
}

int main( int argc, char * argv[] )
{
    if (argc < 3) {
        std::cout << "Invalid arguments" << std::endl;
        exit(-1);
    }
    CMesh mesh,mesh2;
    if (vcg::tri::io::Importer<CMesh>::Open(mesh, argv[1])  != 0)
        std::cout << "Invalid filename" << std::endl;
    
    if (vcg::tri::io::Importer<CMesh>::Open(mesh2, argv[2])  != 0)
        std::cout << "Invalid filename" << std::endl;
    
    std::cout << "Mesh BBox diagonal: " << mesh.bbox.Diag() << std::endl;
    std::cout << "Max point random offset: " << mesh.bbox.Diag() / 1000.0f << std::endl << std::endl;
    
    ComputeAverageVertexDistance(mesh);

    ComputeDistanceBetweenMeshes(mesh, mesh2);

    // map the quality of the vertexes into the color
    tri::UpdateColor<CMesh>::PerVertexQualityRamp(mesh);

    //save the mesh with the average distance as quality and with the per vertex color
    tri::io::ExporterPLY<CMesh>::Save(mesh, "out.ply", tri::io::Mask::IOM_VERTQUALITY | tri::io::Mask::IOM_VERTCOLOR);
}
